package org.eclipse.jdt.internal.compiler.lookup;

import java.util.Set;
import java.util.Collections;
import java.util.HashSet;
import java.util.Collection;
import org.eclipse.jdt.internal.compiler.ast.ExpressionContext;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import java.util.List;
import org.eclipse.jdt.internal.compiler.ast.MessageSend;
import java.util.ArrayList;
import org.eclipse.jdt.internal.compiler.ast.ReferenceExpression;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.ConditionalExpression;
import org.eclipse.jdt.internal.compiler.ast.Invocation;
import org.eclipse.jdt.internal.compiler.ast.Expression;

class ConstraintExpressionFormula extends ConstraintFormula
{
    Expression left;
    boolean isSoft;
    
    ConstraintExpressionFormula(Expression expression, TypeBinding type, int relation) {
        this.left = expression;
        this.right = type;
        this.relation = relation;
    }
    
    ConstraintExpressionFormula(Expression expression, TypeBinding type, int relation, boolean isSoft) {
        this(expression, type, relation);
        this.isSoft = isSoft;
    }
    
    @Override
    public Object reduce(InferenceContext18 inferenceContext) throws InferenceFailureException {
        if (this.relation == 8) {
            return this.left.isPotentiallyCompatibleWith(this.right, inferenceContext.scope) ? ConstraintExpressionFormula.TRUE : ConstraintExpressionFormula.FALSE;
        }
        if (this.right.isProperType(true)) {
            return (this.left.isCompatibleWith(this.right, inferenceContext.scope) || this.left.isBoxingCompatibleWith(this.right, inferenceContext.scope)) ? ConstraintExpressionFormula.TRUE : ConstraintExpressionFormula.FALSE;
        }
        if (!this.canBePolyExpression(this.left)) {
            TypeBinding exprType3 = this.left.resolvedType;
            if (exprType3 == null || !exprType3.isValidBinding()) {
                if (this.left instanceof MessageSend && ((MessageSend)this.left).actualReceiverType instanceof InferenceVariable) {
                    return null;
                }
                return ConstraintExpressionFormula.FALSE;
            }
            else {
                return ConstraintTypeFormula.create(exprType3, this.right, 1, this.isSoft);
            }
        }
        else {
            if ((this.left instanceof Invocation)) {
                Invocation invocation = (Invocation)this.left;
                MethodBinding previousMethod = invocation.binding();
                if (previousMethod == null) {
                    return null;
                }
                MethodBinding method = previousMethod;
                method = previousMethod.shallowOriginal();
                InferenceContext18.SuspendedInferenceRecord prevInvocation = inferenceContext.enterPolyInvocation(invocation, invocation.arguments());
                try {
                    Expression[] arguments = invocation.arguments();
                    TypeBinding[] argumentTypes = (arguments == null) ? Binding.NO_PARAMETERS : new TypeBinding[arguments.length];
                    for (int i = 0; i < argumentTypes.length; ++i) {
                        argumentTypes[i] = arguments[i].resolvedType;
                    }
                    if ((previousMethod instanceof ParameterizedGenericMethodBinding)) {
                        InferenceContext18 innerCtx = invocation.getInferenceContext((ParameterizedGenericMethodBinding)previousMethod);
                        if (innerCtx == null) {
                            TypeBinding exprType = this.left.resolvedType;
                            if (exprType == null || !exprType.isValidBinding()) {
                                return ConstraintExpressionFormula.FALSE;
                            }
                            return ConstraintTypeFormula.create(exprType, this.right, 1, this.isSoft);
                        }
                        else if (innerCtx.stepCompleted >= 1) {
                            inferenceContext.integrateInnerInferenceB2(innerCtx);
                        }
                        else {
                            return ConstraintExpressionFormula.FALSE;
                        }
                    }
                    else {
                        inferenceContext.inferenceKind = inferenceContext.getInferenceKind(previousMethod, argumentTypes);
                        boolean isDiamond = method.isConstructor() && this.left.isPolyExpression(method);
                        inferInvocationApplicability(inferenceContext, method, argumentTypes, isDiamond, inferenceContext.inferenceKind);
                    }
                    if (!inferenceContext.computeB3(invocation, this.right, method)) {
                        return ConstraintExpressionFormula.FALSE;
                    }
                    return null;
                }
                finally {
                    inferenceContext.resumeSuspendedInference(prevInvocation);
                }
            }
            if ((this.left instanceof ConditionalExpression)) {
                ConditionalExpression conditional = (ConditionalExpression)this.left;
                return new ConstraintFormula[] { new ConstraintExpressionFormula(conditional.valueIfTrue, this.right, this.relation, this.isSoft), new ConstraintExpressionFormula(conditional.valueIfFalse, this.right, this.relation, this.isSoft) };
            }
            if ((this.left instanceof LambdaExpression)) {
                LambdaExpression lambda = (LambdaExpression)this.left;
                BlockScope scope = lambda.enclosingScope;
                if (!this.right.isFunctionalInterface(scope)) {
                    return ConstraintExpressionFormula.FALSE;
                }
                ReferenceBinding t = (ReferenceBinding)this.right;
                ParameterizedTypeBinding withWildCards = InferenceContext18.parameterizedWithWildcard(t);
                if (withWildCards != null) {
                    t = findGroundTargetType(inferenceContext, scope, lambda, withWildCards);
                }
                if (t == null) {
                    return ConstraintExpressionFormula.FALSE;
                }
                MethodBinding functionType = t.getSingleAbstractMethod(scope, true);
                if (functionType == null) {
                    return ConstraintExpressionFormula.FALSE;
                }
                TypeBinding[] parameters = functionType.parameters;
                if (parameters.length != lambda.arguments().length) {
                    return ConstraintExpressionFormula.FALSE;
                }
                if (lambda.argumentsTypeElided()) {
                    int i = 0;
                    while (i < parameters.length) {
                        if (!parameters[i].isProperType(true)) {
                            return ConstraintExpressionFormula.FALSE;
                        }
                        ++i;
                    }
                }
                lambda = lambda.resolveExpressionExpecting(t, inferenceContext.scope, inferenceContext);
                if (lambda == null) {
                    return ConstraintExpressionFormula.FALSE;
                }
                if (functionType.returnType == TypeBinding.VOID) {
                    if (!lambda.isVoidCompatible()) {
                        return ConstraintExpressionFormula.FALSE;
                    }
                }
                else if (!lambda.isValueCompatible()) {
                    return ConstraintExpressionFormula.FALSE;
                }
                List<ConstraintFormula> result = new ArrayList();
                if (!lambda.argumentsTypeElided()) {
                    Argument[] arguments2 = lambda.arguments();
                    for (int j = 0; j < parameters.length; ++j) {
                        result.add(ConstraintTypeFormula.create(parameters[j], arguments2[j].type.resolvedType, 4));
                    }
                    if (lambda.resolvedType != null) {
                        result.add(ConstraintTypeFormula.create(lambda.resolvedType, this.right, 2));
                    }
                }
                if (functionType.returnType != TypeBinding.VOID) {
                    TypeBinding r = functionType.returnType;
                    Expression[] exprs = lambda.resultExpressions();
                    for (int k = 0, length = (exprs == null) ? 0 : exprs.length; k < length; ++k) {
                        Expression expr = exprs[k];
                        if (r.isProperType(true) && expr.resolvedType != null) {
                            TypeBinding exprType2 = expr.resolvedType;
                            if (!expr.isConstantValueOfTypeAssignableToType(exprType2, r) && !exprType2.isCompatibleWith(r) && !expr.isBoxingCompatible(exprType2, r, expr, scope)) {
                                return ConstraintExpressionFormula.FALSE;
                            }
                        }
                        else {
                            result.add(new ConstraintExpressionFormula(expr, r, 1, this.isSoft));
                        }
                    }
                }
                if (result.size() == 0) {
                    return ConstraintExpressionFormula.TRUE;
                }
                return result.toArray(new ConstraintFormula[result.size()]);
            }
            else {
                if ((this.left instanceof ReferenceExpression)) {
                    return this.reduceReferenceExpressionCompatibility((ReferenceExpression)this.left, inferenceContext);
                }
                return ConstraintExpressionFormula.FALSE;
            }
        }
    }
    
    public static ReferenceBinding findGroundTargetType(InferenceContext18 inferenceContext, BlockScope scope, LambdaExpression lambda, ParameterizedTypeBinding targetTypeWithWildCards) {
        if (lambda.argumentsTypeElided()) {
            return lambda.findGroundTargetTypeForElidedLambda(scope, targetTypeWithWildCards);
        }
        InferenceContext18.SuspendedInferenceRecord previous = inferenceContext.enterLambda(lambda);
        try {
            return inferenceContext.inferFunctionalInterfaceParameterization(lambda, scope, targetTypeWithWildCards);
        }
        finally {
            inferenceContext.resumeSuspendedInference(previous);
        }
    }
    
    private boolean canBePolyExpression(Expression expr) {
        ExpressionContext previousExpressionContext = expr.getExpressionContext();
        if (previousExpressionContext == ExpressionContext.VANILLA_CONTEXT) {
            this.left.setExpressionContext(ExpressionContext.ASSIGNMENT_CONTEXT);
        }
        try {
            return expr.isPolyExpression();
        }
        finally {
            expr.setExpressionContext(previousExpressionContext);
        }
    }
    
    private Object reduceReferenceExpressionCompatibility(ReferenceExpression reference, InferenceContext18 inferenceContext) {
        TypeBinding t = this.right;
        if (t.isProperType(true)) {
            throw new IllegalStateException("Should not reach here with T being a proper type");
        }
        if (!t.isFunctionalInterface(inferenceContext.scope)) {
            return ConstraintExpressionFormula.FALSE;
        }
        MethodBinding functionType = t.getSingleAbstractMethod(inferenceContext.scope, true);
        if (functionType == null) {
            return ConstraintExpressionFormula.FALSE;
        }
        reference = reference.resolveExpressionExpecting(t, inferenceContext.scope, inferenceContext);
        MethodBinding potentiallyApplicable = (reference != null) ? reference.binding : null;
        if (potentiallyApplicable == null) {
            return ConstraintExpressionFormula.FALSE;
        }
        if (reference.isExactMethodReference()) {
            List<ConstraintFormula> newConstraints = new ArrayList();
            TypeBinding[] p = functionType.parameters;
            int n2 = p.length;
            TypeBinding[] pPrime = potentiallyApplicable.parameters;
            int k = pPrime.length;
            int offset = 0;
            if (n2 == k + 1) {
                newConstraints.add(ConstraintTypeFormula.create(p[0], reference.lhs.resolvedType, 1));
                offset = 1;
            }
            for (int j = offset; j < n2; ++j) {
                newConstraints.add(ConstraintTypeFormula.create(p[j], pPrime[j - offset], 1));
            }
            TypeBinding r2 = functionType.returnType;
            if (r2 != TypeBinding.VOID) {
                TypeBinding rAppl = (potentiallyApplicable.isConstructor() && !reference.isArrayConstructorReference()) ? potentiallyApplicable.declaringClass : potentiallyApplicable.returnType;
                if (rAppl == TypeBinding.VOID) {
                    return ConstraintExpressionFormula.FALSE;
                }
                TypeBinding rPrime2 = rAppl.capture(inferenceContext.scope, reference.sourceStart, reference.sourceEnd);
                newConstraints.add(ConstraintTypeFormula.create(rPrime2, r2, 1));
            }
            return newConstraints.toArray(new ConstraintFormula[newConstraints.size()]);
        }
        else {
            int n = functionType.parameters.length;
            int i = 0;
            while (i < n) {
                if (!functionType.parameters[i].isProperType(true)) {
                    return ConstraintExpressionFormula.FALSE;
                }
                ++i;
            }
            MethodBinding compileTimeDecl = potentiallyApplicable;
            if (!compileTimeDecl.isValidBinding()) {
                return ConstraintExpressionFormula.FALSE;
            }
            TypeBinding r = functionType.isConstructor() ? functionType.declaringClass : functionType.returnType;
            if (r.id == 6) {
                return ConstraintExpressionFormula.TRUE;
            }
            MethodBinding original = compileTimeDecl.shallowOriginal();
            TypeBinding compileTypeReturn = original.isConstructor() ? original.declaringClass : original.returnType;
            if (reference.typeArguments == null && ((original.typeVariables() != Binding.NO_TYPE_VARIABLES && compileTypeReturn.mentionsAny(original.typeVariables(), -1)) || (original.isConstructor() && compileTimeDecl.declaringClass.isRawType()))) {
                TypeBinding[] argumentTypes;
                if (t.isParameterizedType()) {
                    MethodBinding capturedFunctionType = ((ParameterizedTypeBinding)t).getSingleAbstractMethod(inferenceContext.scope, true, reference.sourceStart, reference.sourceEnd);
                    argumentTypes = capturedFunctionType.parameters;
                }
                else {
                    argumentTypes = functionType.parameters;
                }
                InferenceContext18.SuspendedInferenceRecord prevInvocation = inferenceContext.enterPolyInvocation(reference, reference.createPseudoExpressions(argumentTypes));
                try {
                    InferenceContext18 innerContex = reference.getInferenceContext((ParameterizedMethodBinding)compileTimeDecl);
                    int innerInferenceKind = (innerContex != null) ? innerContex.inferenceKind : 1;
                    inferInvocationApplicability(inferenceContext, original, argumentTypes, original.isConstructor(), innerInferenceKind);
                    if (!inferenceContext.computeB3(reference, r, original)) {
                        return ConstraintExpressionFormula.FALSE;
                    }
                    if (!original.isConstructor() || reference.receiverType.isRawType() || reference.receiverType.typeArguments() == null) {
                        return null;
                    }
                }
                catch (InferenceFailureException ex) {
                    return ConstraintExpressionFormula.FALSE;
                }
                finally {
                    inferenceContext.resumeSuspendedInference(prevInvocation);
                }
            }
            TypeBinding rPrime = compileTimeDecl.isConstructor() ? compileTimeDecl.declaringClass : compileTimeDecl.returnType.capture(inferenceContext.scope, reference.sourceStart(), reference.sourceEnd());
            if (rPrime.id == 6) {
                return ConstraintExpressionFormula.FALSE;
            }
            return ConstraintTypeFormula.create(rPrime, r, 1, this.isSoft);
        }
    }
    
    static void inferInvocationApplicability(InferenceContext18 inferenceContext, MethodBinding method, TypeBinding[] arguments, boolean isDiamond, int checkType) {
        TypeVariableBinding[] typeVariables = method.getAllTypeVariables(isDiamond);
        InferenceVariable[] inferenceVariables = inferenceContext.createInitialBoundSet(typeVariables);
        int paramLength = method.parameters.length;
        TypeBinding varArgsType = null;
        if (method.isVarargs()) {
            int varArgPos = paramLength - 1;
            varArgsType = method.parameters[varArgPos];
        }
        inferenceContext.createInitialConstraintsForParameters(method.parameters, checkType == 3, varArgsType, method);
        inferenceContext.addThrowsContraints(typeVariables, inferenceVariables, method.thrownExceptions);
    }
    
    static boolean inferPolyInvocationType(InferenceContext18 inferenceContext, InvocationSite invocationSite, TypeBinding targetType, MethodBinding method) throws InferenceFailureException {
        TypeBinding[] typeArguments = invocationSite.genericTypeArguments();
        if (typeArguments == null) {
            TypeBinding returnType = method.isConstructor() ? method.declaringClass : method.returnType;
            if (returnType == TypeBinding.VOID) {
                throw new InferenceFailureException("expression has no value");
            }
            if (inferenceContext.usesUncheckedConversion) {
                TypeBinding erasure = inferenceContext.environment.convertToRawType(returnType, false);
                ConstraintTypeFormula newConstraint4 = ConstraintTypeFormula.create(erasure, targetType, 1);
                return inferenceContext.reduceAndIncorporate(newConstraint4);
            }
            TypeBinding rTheta = inferenceContext.substitute(returnType);
            ParameterizedTypeBinding parameterizedType = InferenceContext18.parameterizedWithWildcard(rTheta);
            if (parameterizedType != null && parameterizedType.arguments != null) {
                TypeBinding[] arguments = parameterizedType.arguments;
                InferenceVariable[] betas = inferenceContext.addTypeVariableSubstitutions(arguments);
                ParameterizedTypeBinding gbeta = inferenceContext.environment.createParameterizedType(parameterizedType.genericType(), betas, parameterizedType.enclosingType(), parameterizedType.getTypeAnnotations());
                inferenceContext.currentBounds.captures.put(gbeta, parameterizedType);
                parameterizedType = parameterizedType.capture(inferenceContext.scope, invocationSite.sourceStart(), invocationSite.sourceEnd());
                arguments = parameterizedType.arguments;
                for (int i = 0, length = arguments.length; i < length; ++i) {
                    if (arguments[i].isCapture() && arguments[i].isProperType(true)) {
                        CaptureBinding capture = (CaptureBinding)arguments[i];
                        inferenceContext.currentBounds.addBound(new TypeBound(betas[i], capture, 4), inferenceContext.environment);
                    }
                }
                ConstraintTypeFormula newConstraint3 = ConstraintTypeFormula.create(gbeta, targetType, 1);
                return inferenceContext.reduceAndIncorporate(newConstraint3);
            }
            if ((rTheta.leafComponentType() instanceof InferenceVariable)) {
                InferenceVariable alpha = (InferenceVariable)rTheta.leafComponentType();
                TypeBinding targetLeafType = targetType.leafComponentType();
                boolean toResolve = false;
                if (inferenceContext.currentBounds.condition18_5_2_bullet_3_3_1(alpha, targetLeafType)) {
                    toResolve = true;
                }
                else if (inferenceContext.currentBounds.condition18_5_2_bullet_3_3_2(alpha, targetLeafType, inferenceContext)) {
                    toResolve = true;
                }
                else if (targetLeafType.isPrimitiveType()) {
                    TypeBinding wrapper = inferenceContext.currentBounds.findWrapperTypeBound(alpha);
                    if (wrapper != null) {
                        toResolve = true;
                    }
                }
                if (toResolve) {
                    BoundSet solution = inferenceContext.solve(new InferenceVariable[] { alpha });
                    if (solution == null) {
                        return false;
                    }
                    TypeBinding u = solution.getInstantiation(alpha, null).capture(inferenceContext.scope, invocationSite.sourceStart(), invocationSite.sourceEnd());
                    if (rTheta.dimensions() != 0) {
                        u = inferenceContext.environment.createArrayType(u, rTheta.dimensions());
                    }
                    ConstraintTypeFormula newConstraint = ConstraintTypeFormula.create(u, targetType, 1);
                    return inferenceContext.reduceAndIncorporate(newConstraint);
                }
            }
            ConstraintTypeFormula newConstraint2 = ConstraintTypeFormula.create(rTheta, targetType, 1);
            if (!inferenceContext.reduceAndIncorporate(newConstraint2)) {
                return false;
            }
        }
        return true;
    }
    
    @Override
    Collection<InferenceVariable> inputVariables(InferenceContext18 context) {
        if ((this.left instanceof LambdaExpression)) {
            if ((this.right instanceof InferenceVariable)) {
                return Collections.singletonList((InferenceVariable)this.right);
            }
            if (this.right.isFunctionalInterface(context.scope)) {
                LambdaExpression lambda = (LambdaExpression)this.left;
                ReferenceBinding targetType = (ReferenceBinding)this.right;
                ParameterizedTypeBinding withWildCards = InferenceContext18.parameterizedWithWildcard(targetType);
                if (withWildCards != null) {
                    targetType = findGroundTargetType(context, lambda.enclosingScope, lambda, withWildCards);
                }
                if (targetType == null) {
                    return ConstraintExpressionFormula.EMPTY_VARIABLE_LIST;
                }
                MethodBinding sam2 = targetType.getSingleAbstractMethod(context.scope, true);
                Set<InferenceVariable> variables2 = new HashSet();
                if (lambda.argumentsTypeElided()) {
                    for (int len2 = sam2.parameters.length, j = 0; j < len2; ++j) {
                        sam2.parameters[j].collectInferenceVariables(variables2);
                    }
                }
                if (sam2.returnType != TypeBinding.VOID) {
                    TypeBinding r = sam2.returnType;
                    LambdaExpression resolved = lambda.resolveExpressionExpecting(this.right, context.scope, context);
                    Expression[] resultExpressions = (resolved != null) ? resolved.resultExpressions() : null;
                    for (int k = 0, length = (resultExpressions == null) ? 0 : resultExpressions.length; k < length; ++k) {
                        variables2.addAll(new ConstraintExpressionFormula(resultExpressions[k], r, 1).inputVariables(context));
                    }
                    return variables2;
                }
                return variables2;
            }
        }
        else if ((this.left instanceof ReferenceExpression)) {
            if ((this.right instanceof InferenceVariable)) {
                return Collections.singletonList((InferenceVariable)this.right);
            }
            if (this.right.isFunctionalInterface(context.scope) && !this.left.isExactMethodReference()) {
                MethodBinding sam = this.right.getSingleAbstractMethod(context.scope, true);
                Set<InferenceVariable> variables = new HashSet();
                for (int len = sam.parameters.length, i = 0; i < len; ++i) {
                    sam.parameters[i].collectInferenceVariables(variables);
                }
                return variables;
            }
        }
        else if (this.left instanceof ConditionalExpression && this.left.isPolyExpression()) {
            ConditionalExpression expr = (ConditionalExpression)this.left;
            Set<InferenceVariable> variables = new HashSet();
            variables.addAll(new ConstraintExpressionFormula(expr.valueIfTrue, this.right, 1).inputVariables(context));
            variables.addAll(new ConstraintExpressionFormula(expr.valueIfFalse, this.right, 1).inputVariables(context));
            return variables;
        }
        return ConstraintExpressionFormula.EMPTY_VARIABLE_LIST;
    }
    
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer().append('⟨');
        this.left.printExpression(4, buf);
        buf.append(ReductionResult.relationToString(this.relation));
        this.appendTypeName(buf, this.right);
        buf.append('⟩');
        return buf.toString();
    }
}
